"""
This module defines the methods for social community detection.
"""


def get_groups(
        a_graph,
        method='component_infomap', return_form='membership'):
    """
    Return the grouping of the provided graph object using the specified
    method. The grouping is returned as a list of sets each holding all
    members of a group.

    Parameters
    ==========
    a_graph: :class:`igraph.Graph`
        The graph to partition
    method: str (default='component_infomap')
        String specifying which method to use. If two methods
        should be used one after the other they should be separated by `_`.
        Default: 'component_infomap' which will first consider all
        disconnected components as groups then apply infomap on all of
        those groups to optionally further split.
    return_form: str (default='membership')
        Determines the format of how the social group structure should be
        returned. Options are:

        * ``'membership'``: A list returning for each `index` node the group it
          belongs to.
        * ``'memberlists'``: Dictionary with a list of members `value` for each
          group `key`.

    Returns
    =======
    dict
      Depending on what was chosen for the `return_form` attribute, either the
      membership dict, i.e.::

          {
              node_id: group_id,
              ...
          }

      or the memberlist dict, i.e.::

          {
              group_id: [node1_id, node2_id, ...],
              ...
          }

      (value) is returned.

    """
    # methods = method.split('_')
    # For now only 'component_infomap' is allowed as procedure
    if method == 'component_infomap':
        # first the connected components
        a_graph.vs['component'] = a_graph.clusters(
                ).membership
        components = set(a_graph.vs['component'])
        # create for each component a graph and apply infomap to it
        node_membership = {}
        # print(
        #     'INFO: Found {0} disconnected components'.format(len(components))
        # )
        if components:
            # do the community detection on each component and create a
            # compound group id: component_group
            for component in components:
                _comp_graph = a_graph.subgraph(
                    [
                        node['name']
                        for node in a_graph.vs
                        if node['component'] == component
                        ]
                    )
                _infompa_comp_graph = _comp_graph.community_infomap('weight')
                _comp_graph.vs['_group'] = _infompa_comp_graph.membership
                for node in _comp_graph.vs:
                    node_membership[node['name']] = '{0}_{1}'.format(
                            node['component'], node['_group']
                            )
                del _infompa_comp_graph
        else:
            _infompa_comp_graph = a_graph.community_infomap('weight')
            a_graph.vs['group'] = _infompa_comp_graph.membership
            node_membership = {
                node['name']: node['group']
                for node in a_graph.vs
            }
        group_membership = {}
        for node in node_membership:
            try:
                group_membership[node_membership[node]].append(node)
            except KeyError:
                group_membership[node_membership[node]] = [node]
    if return_form == 'membership':
        # nbr_nodes = len(a_graph.vs['name'])
        # membership = [None]*nbr_nodes
        # for g, members in group_membership.items():
        #     for member in members:
        #         membership[member] = g
        # return membership
        return node_membership
    elif return_form == 'memberlists':
        # return [_group for _group in group_membership.values()]
        return group_membership
    else:
        return None



"""
Module to define the spatial based partitioning
"""
from sklearn.cluster import KMeans


def get_groups(
        node_locations, nbr_clusters,
        return_form='membership', random_state=None,
        **kwargs
        ):
    """
    Perform k-means clustering on the provided node locations.

    Parameters
    ==========
    node_locations: dict
      Specify for each node (key) its position (value) in the form of a tuple,
      `(x, y)`.
    nbr_clusters: int
      The number of clusters to find.
    return_form: str (default='membership')
        Determines the format of how the social group structure should be
        returned. Options are:

        * ``'membership'``: A `dict` returning for each node (key) the group it
          belongs to (value).
        * ``'memberlists'``: Dictionary with a list of members (value) for each
          group (key).
    random_state : int, RandomState instance, default=None
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
    **kwargs
      All keyword arguments forwarded to :class:`sklearn.cluster.KMeans`.


    Returns
    =======
    dict
      Depending on what was chosen for the `return_form` attribute, either the
      membership dict, i.e.::

          {
              node_id: group_id,
              ...
          }

      or the memberlist dict, i.e.::

          {
              group_id: [node1_id, node2_id, ...],
              ...
          }

      (value) is returned.

    """
    nodes = sorted(node_locations.keys())
    positions = [node_locations[n] for n in nodes]
    membership_predict = KMeans(
        n_clusters=nbr_clusters,
        random_state=random_state,
        **kwargs
    ).fit_predict(positions)
    group_membership = {
            node: membership_predict[i]
            for i, node in enumerate(nodes)
        }
    # note: could use default dict here
    kmeans_grouping = {i: [] for i in range(nbr_clusters)}
    for i, memb in enumerate(membership_predict):
        kmeans_grouping[memb].append(nodes[i])
    if return_form == 'membership':
        membership = [None]*len(nodes)
        for g, members in kmeans_grouping.items():
            for member in members:
                membership[member] = g
        return membership
        return group_membership
    elif return_form == 'memberlists':
        return kmeans_grouping
    else:
        return None



"""
Module holding functions for a quantitative comparison of groupings
"""
from sklearn import metrics

from .spatial import get_groups as get_spatial_groups


def social_vs_spatial(node_locations, soc_membership, **kwargs):
    """

    This method relies on the
    :func:`~.sklearn.metrics.homogeneity_completeness_v_measure` method form
    the :obj:`.sklearn.metrics` package.

    .. seealso::

      | V-Measure:
      | A conditional entropy-based external cluster evaluation measure
      | Andrew Rosenberg and Julia Hirschberg, 2007
      | http://acl.ldc.upenn.edu/D/D07/D07-1043.pdf

    Parameters
    ==========
    node_locations: dict(int, tuple)
      Specify for each node (key) its position (value) in the form of a (x, y)
      tuple.
    soc_membership: dict
      Holds for each node (key) the social group affiliation (value).
    **kwargs
      All keyword arguments forwarded to :func:`.spatial.get_groups`

    Returns
    =======
    homogeneity, completeness, v-measure: tuple
      Homogeneity, completeness and v-measure as computed by the
      :func:`~.sklearn.metrics.homogeneity_completeness_v_measure`.
    """
    nbr_clusters = len(set(soc_membership.values()))
    spatial_membership = get_spatial_groups(
            node_locations, nbr_clusters, **kwargs
        )
    nodes = sorted(soc_membership.keys())
    # create lists for both memberships with the same ordering
    _soc_membership = [soc_membership[node] for node in nodes]
    _spat_membership = [spatial_membership[node] for node in nodes]
    h, c, v = metrics.homogeneity_completeness_v_measure(
            _soc_membership, _spat_membership
    )
    return h, c, v



"""
The module exposes the functions :func:`get_social_groups` and
:func:`compare_soc_spat`.

:func:`get_social_groups` is defined in :mod:`.sospcat.social` as
:func:`.spspcat.social.get_groups`.

:func:`compare_soc_spat` is defined in :mod:`.sospcat.compare` as
:func:`.spspcat.social.social_vs_spatial`.

.. moduleauthor:: J. I. Liechti
"""

from ._version import __version__  # noqa: F401

from .social import get_groups as get_social_groups  # noqa: F401
from .compare import social_vs_spatial as compare_soc_spat  # noqa: F401

__all__ = ['get_social_groups', 'compare_soc_spat']



#!/usr/bin/env python
import codecs
import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))


def get_version_and_cmdclass(package_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location(
            "version", os.path.join(package_path, "_version.py")
        )
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


def find_packages():
    """adapted from IPython's setupbase.find_packages()"""
    packages = []
    for dir, subdirs, files in os.walk('sospcat'):
        package = dir.replace(os.path.sep, '.')
        if '__init__.py' not in files:
            # not a package
            continue
        # if sys.version_info < (3, 4)\
        #        and 'asyncio' in package and 'sdist' not in sys.argv:
        #     # Don't install asyncio packages on old Python
        #     # avoids issues with tools like compileall, pytest, etc.
        #     # that get confused by presence of Python 3-only sources,
        #     # even when they are never imported.
        #     continue
        packages.append(package)
    return packages


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def install_requires(req_file='requirements.txt'):
    with open(req_file, 'r') as fobj:
        requires = fobj.read().splitlines()
    # if sys.version_info > (3, 4):
    #     requires.extend(['websockets'])
    return requires


def get_readme(name='README.md'):
    r"""Return the content of the README file.

    By default it is assumed that the README file is in the same directory as
    setup.py.
    The README format is determined via the file extension and should be either
    `.md` or `.rst`. Any other format will lead to the content being processed
    as plain text.
    """
    with open(os.path.join(here, name), 'r') as fobj:
        readme_content = fobj.read()
    if name.endswith('.md'):
        long_description_content_type = 'text/markdown'
    elif name.endswith('.rst'):
        long_description_content_type = 'text/x-rst'
    else:
        long_description_content_type = 'text/plain'

    return readme_content, long_description_content_type


version, cmdclass = get_version_and_cmdclass("sospcat")
long_description, long_description_content_type = get_readme()


setup(
    name='SoSpCAT',
    version=version,
    cmdclass=cmdclass,
    packages=find_packages(),
    description='Social-Spatial Community Assignment Test',
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    url='https://SoSpcATPy.readthedocs.io',
    project_urls={
            "Bug Tracker": "https://github.com/j-i-l/SoSpCATpy/issues",
            "Documentation": 'https://sospcatpy.readthedocs.io',
            "Source Code": "https://github.com/j-i-l/SoSpCATpy",
        },
    author='Jonas I. Liechti [aut, cre]',
    author_email='jonas.i.liechti@gmail.com',
    license='BSD',
    install_requires=install_requires(),
    keywords='social spatial group cluster groupstructure',
    classifiers=[
          'Intended Audience :: Science/Research',
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python',
          'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    # include_package_data=True,  # includes whatever is in MANIFEST.in
    # test_suite='tests',
)
